import asyncio
from ipaddress import ip_address

import typer
from rich import print

from .main import (
    trydan_charging,
    trydan_connected,
    trydan_intensity,
    trydan_lock,
    trydan_pause,
    trydan_ready,
    trydan_resume,
    trydan_set,
    trydan_status,
    trydan_unlock,
)

app = typer.Typer()


@app.command()
def status(ip: str) -> None:
    """Retrieve Trydan Status."""
    print("Connecting to %s", ip_address(ip))

    asyncio.run(trydan_status(ip))


@app.command()
def connected(ip: str) -> None:
    """Retrieve Trydan Status."""
    print("Connecting to %s", ip_address(ip))

    asyncio.run(trydan_connected(ip))


@app.command()
def charging(ip: str) -> None:
    """Retrieve Trydan Status."""
    print("Connecting to %s", ip_address(ip))

    asyncio.run(trydan_charging(ip))


@app.command()
def ready(ip: str) -> None:
    """Retrieve Trydan Status."""
    print("Connecting to %s", ip_address(ip))

    asyncio.run(trydan_ready(ip))


@app.command()
def pause(ip: str) -> None:
    """Pause Trydan EVSE."""
    print("Connecting to %s", ip_address(ip))

    asyncio.run(trydan_pause(ip))


@app.command()
def resume(ip: str) -> None:
    """Resume Trydan EVSE."""
    print("Connecting to %s", ip_address(ip))

    asyncio.run(trydan_resume(ip))


@app.command()
def lock(ip: str) -> None:
    """Lock Trydan EVSE."""
    print("Connecting to %s", ip_address(ip))

    asyncio.run(trydan_lock(ip))


@app.command()
def unlock(ip: str) -> None:
    """Unlock Trydan EVSE."""
    print("Connecting to %s", ip_address(ip))

    asyncio.run(trydan_unlock(ip))


@app.command()
def intensity(ip: str, intensity: int) -> None:
    """Set Intensity in Trydan EVSE."""
    print("Connecting to %s", ip_address(ip))

    asyncio.run(trydan_intensity(ip, intensity))


@app.command()
def set(ip: str, keyword: str, value: str) -> None:
    """Set KeyWord value in Trydan."""
    print("Connecting to %s", ip_address(ip))

    asyncio.run(trydan_set(ip, keyword, value))



import logging
from http import HTTPStatus
from typing import Any

import httpx
import orjson
from httpcore import ConnectTimeout
from tenacity import retry, retry_if_exception_type, wait_random_exponential

from .const import API_TIMEOUT, KEYWORDS
from .exceptions import (
    TrydanCommunicationError,
    TrydanInvalidKeyword,
    TrydanInvalidResponse,
    TrydanInvalidValue,
    TrydanRetryLater,
)
from .models.trydan import (
    ChargePointTimerState,
    ChargeState,
    DynamicPowerMode,
    DynamicState,
    LockState,
    PauseDynamicState,
    PauseState,
    ReadyState,
    TrydanData,
)

VALIDATION = {
    "ChargeState": lambda x: x in ChargePointTimerState,
    "DynamicPowerMode": lambda x: x in DynamicPowerMode,
    "Dynamic": lambda x: x in DynamicState,
    "Locked": lambda x: x in LockState,
    "PauseDynamic": lambda x: x in PauseDynamicState,
    "Paused": lambda x: x in PauseState,
    "Intensity": lambda x: x >= 6 and x <= 32,
    "MinIntensity": lambda x: x >= 6 and x <= 32,
    "MaxIntensity": lambda x: x >= 6 and x <= 32,
}

_LOGGER = logging.getLogger(__name__)


class Trydan:
    """Class for communicating with Trydan."""

    def __init__(
        self,
        host: str,
        client: httpx.AsyncClient | None = None,
    ) -> None:
        """Initialize."""
        self._host = host
        self._client = client or httpx.AsyncClient()
        self._timeout = API_TIMEOUT
        self._data: TrydanData | None = None

    @retry(
        retry=retry_if_exception_type(
            (
                httpx.NetworkError,
                httpx.TimeoutException,
                httpx.RemoteProtocolError,
            )
        ),
        wait=wait_random_exponential(multiplier=2, max=3),
    )
    async def request(self, endpoint: str) -> httpx.Response:
        """Make a request to Trydan."""
        return await self._request(endpoint)

    async def _request(
        self,
        url: str,
    ) -> httpx.Response:
        """Make a request to Trydan."""
        _LOGGER.debug("Requesting %s with timeout %s", url, self._timeout)
        response = await self._client.get(
            url,
            timeout=self._timeout,
        )

        self.raw_data = {
            "content": response.content,
            "status_code": response.status_code,
        }

        status_code = response.status_code
        if status_code in (HTTPStatus.UNAUTHORIZED, HTTPStatus.FORBIDDEN):
            raise TrydanCommunicationError(
                f"Failed for {url} with status {status_code}"
            )
        if status_code != HTTPStatus.OK:
            raise TrydanInvalidResponse(f"Failed for {url} with status {status_code}")

        return response

    async def _json_request(self, end_point: str) -> Any:
        """Make a request to Trydan and return the JSON response."""
        response = await self._request(end_point)
        try:
            return orjson.loads(response.content)
        except orjson.JSONDecodeError as err:
            _LOGGER.error(
                "Error decoding JSON response from Trydan: ", response.content
            )
            raise TrydanInvalidResponse(
                "Error decoding JSON response from Trydan"
            ) from err

    async def get_data(self) -> TrydanData:
        """Get data from Trydan."""
        try:
            data = await self._json_request(f"http://{self._host}/RealTimeData")
        except (ConnectTimeout, httpx.ConnectTimeout) as err:
            raise TrydanRetryLater("Timeout connecting to Trydan") from err
        except httpx.ReadTimeout as err:
            raise TrydanRetryLater("Timeout reading from Trydan") from err

        self._data = TrydanData.from_api(data)
        return self._data

    async def set_keyword(
        self,
        keyword: str,
        value: str
        | int
        | ChargePointTimerState
        | DynamicPowerMode
        | DynamicState
        | LockState
        | PauseDynamicState
        | PauseState
        | TrydanData,
    ) -> None:
        """Set a keyword in Trydan."""
        if keyword not in KEYWORDS:
            raise TrydanInvalidKeyword(f"Keyword {keyword} is not valid")

        if keyword in VALIDATION:
            if not VALIDATION[keyword](value):
                raise TrydanInvalidValue(
                    f"Value {value} is not valid for keyword {keyword}"
                )

        url = f"http://{self._host}/write/{keyword}={value}"
        _LOGGER.debug("HTTP GET: %s", url)
        try:
            data = await self._request(url)
        except ConnectTimeout as err:
            raise TrydanRetryLater("Timeout connecting to Trydan") from err

        if data.status_code != 200 or data.content != b"OK":
            raise TrydanInvalidValue(
                f"Failed for {keyword}={value}"
                " code={data.status_code} : <{data.content}>"
            )

    @property
    def data(self) -> TrydanData | None:
        """Return cached version of Trydan EVSE."""
        if self._data is None:
            raise TrydanRetryLater("no initial data retrieved")
        return self._data

    @property
    def host(self) -> str:
        """Return the Trydan host."""
        return self._host

    @property
    def id(self) -> str | None:
        """Return the Trydan ID."""
        if self._data is None:
            raise TrydanRetryLater("No data available")
        return self._data.ID

    @property
    def firmware_version(self) -> str | None:
        """Return the Trydan firmware version."""
        if self._data is None:
            raise TrydanRetryLater("No data available")
        return self._data.firmware_version

    @property
    def connected(self) -> bool:
        """Return the Trydan connection state."""
        if self._data is None:
            raise TrydanRetryLater("No data available")
        return self._data.charge_state != ChargeState.NOT_CONNECTED

    @property
    def charging(self) -> bool:
        """Return the Trydan charging state."""
        if self._data is None:
            raise TrydanRetryLater("No data available")
        return self._data.charge_state == ChargeState.CONNECTED_CHARGING

    @property
    def ready(self) -> bool:
        """Return the Trydan ready state."""
        if self._data is None:
            raise TrydanRetryLater("No data available")
        return self._data.ready_state == ReadyState.READY

    async def pause(self, value: bool = True) -> None:
        """Pause state of current charging session."""
        await self.set_keyword(
            "Paused", PauseState.PAUSED if value else PauseState.NOT_PAUSED
        )

    async def resume(self) -> None:
        """Resume state of current charging session."""
        await self.pause(False)

    async def lock(self, value: bool = True) -> None:
        """Disabling state of Charge Point."""
        await self.set_keyword(
            "Locked", LockState.ENABLED if value else LockState.DISABLED
        )

    async def unlock(self) -> None:
        """Disabling state of Charge Point."""
        await self.lock(False)

    async def timer(self, value: bool = True) -> None:
        """Set the Charge Point Timer state."""
        await self.set_keyword(
            "Timer",
            ChargePointTimerState.TIMER_ON
            if value
            else ChargePointTimerState.TIMER_OFF,
        )

    async def timer_disable(self) -> None:
        """Disable the Charge Point Timer."""
        await self.timer(False)

    async def intensity(self, current: int) -> None:
        """Set the intensity of the Charge Point."""
        if not (current >= 6 and current <= 32):
            raise TrydanInvalidValue("Intensity must be between 6 and 32")

        await self.set_keyword("Intensity", current)

    async def dynamic(self, value: bool = True) -> None:
        """Set the Dynamic Intensity Modulation state."""
        await self.set_keyword(
            "Dynamic", DynamicState.ENABLED if value else DynamicState.DISABLED
        )

    async def dynamic_disable(self) -> None:
        """Disable the Dynamic Intensity Modulation."""
        await self.dynamic(False)

    async def min_intensity(self, current: int) -> None:
        """Set the minimum intensity of the Charge Point."""
        if not (current >= 6 and current <= 32):
            raise TrydanInvalidValue("Intensity must be between 6 and 32")

        await self.set_keyword("MinIntensity", current)

    async def max_intensity(self, current: int) -> None:
        """Set the maximum intensity of the Charge Point."""
        if not (current >= 6 and current <= 32):
            raise TrydanInvalidValue("Intensity must be between 6 and 32")

        await self.set_keyword("MaxIntensity", current)

    async def pause_dynamic(self, value: bool = True) -> None:
        """Set the Pause Dynamic state."""
        await self.set_keyword(
            "PauseDynamic",
            PauseDynamicState.NOT_MODULATING if value else PauseDynamicState.MODULATING,
        )

    async def resume_dynamic(self) -> None:
        """Resume the Pause Dynamic state."""
        await self.pause_dynamic(False)

    async def dynamic_power_mode(self, mode: DynamicPowerMode) -> None:
        """Set the Dynamic Power Mode."""
        await self.set_keyword("DynamicPowerMode", mode)

    async def contracted_power(self, power: int) -> None:
        """Set the Contracted Power."""
        if not (power > 0):
            raise TrydanInvalidValue("Contracted Power must be positive")
        await self.set_keyword("ContractedPower", power)



from rich import print

from .trydan import Trydan


async def trydan_status(ip: str) -> int:
    """Retrieve Trydan Status."""
    trydan = Trydan(ip)
    data = await trydan.get_data()

    print(data)

    return 0


async def trydan_connected(ip: str) -> int:
    """Retrieve Trydan Status."""
    trydan = Trydan(ip)
    await trydan.get_data()

    print(trydan.connected)

    return 0


async def trydan_charging(ip: str) -> int:
    """Retrieve Trydan Status."""
    trydan = Trydan(ip)
    await trydan.get_data()

    print(trydan.charging)

    return 0


async def trydan_ready(ip: str) -> int:
    """Retrieve Trydan Status."""
    trydan = Trydan(ip)
    await trydan.get_data()

    print(trydan.ready)

    return 0


async def trydan_set(ip: str, keyword: str, value: str) -> int:
    """Set KeyWord value in Trydan."""
    trydan = Trydan(ip)
    try:
        await trydan.set_keyword(keyword, value)
    except Exception as e:
        print(e)
        return 1
    return 0


async def trydan_pause(ip: str) -> int:
    """Pause Trydan."""
    trydan = Trydan(ip)
    try:
        await trydan.pause()
    except Exception as e:
        print(e)
        return 1
    return 0


async def trydan_resume(ip: str) -> int:
    """Resume Trydan."""
    trydan = Trydan(ip)
    try:
        await trydan.resume()
    except Exception as e:
        print(e)
        return 1
    return 0


async def trydan_lock(ip: str) -> int:
    """Lock Trydan."""
    trydan = Trydan(ip)
    try:
        await trydan.lock()
    except Exception as e:
        print(e)
        return 1

    return 0


async def trydan_unlock(ip: str) -> int:
    """Unlock Trydan."""
    trydan = Trydan(ip)
    try:
        await trydan.unlock()
    except Exception as e:
        print(e)
        return 1
    return 0


async def trydan_intensity(ip: str, intensity: int) -> int:
    """Set Intensity in Trydan."""
    trydan = Trydan(ip)
    try:
        await trydan.intensity(intensity)
    except Exception as e:
        print(e)
        return 1

    return 0



import pytest
import respx
from httpx import Response

from pytrydan.exceptions import (
    TrydanCommunicationError,
    TrydanInvalidResponse,
)

from .conftest import _get_mock_trydan, _load_json_fixture


@pytest.mark.asyncio
@respx.mock
async def test_bad_status():
    respx.get("/RealTimeData").mock(return_value=Response(404, json={}))

    envoy = await _get_mock_trydan()
    with pytest.raises(TrydanInvalidResponse):
        data = await envoy.get_data()
        assert data is None

    respx.get("/RealTimeData").mock(return_value=Response(403, json={}))

    envoy = await _get_mock_trydan()
    with pytest.raises(TrydanCommunicationError):
        data = await envoy.get_data()
        assert data is None


@pytest.mark.asyncio
@respx.mock
async def test_status():
    respx.get("/RealTimeData").mock(
        return_value=Response(200, json=_load_json_fixture("RealTimeData"))
    )

    envoy = await _get_mock_trydan()
    data = await envoy.get_data()
    assert data is not None

    assert data.charge_state == 1
    assert data.ready_state == 1
    assert data.charge_power == 0
    assert data.charge_energy == 7.6
    assert data.slave_error == 0
    assert data.charge_time == 9979
    assert data.house_power == 0
    assert data.fv_power == 0
    assert data.paused == 0
    assert data.locked == 0
    assert data.timer == 1
    assert data.intensity == 16
    assert data.dynamic == 0
    assert data.min_intensity == 6
    assert data.max_intensity == 16
    assert data.pause_dynamic == 0
    assert data.firmware_version == "1.6.18"
    assert data.dynamic_power_mode == 2
    assert data.contracted_power == 4600


@pytest.mark.asyncio
@respx.mock
async def test_status_charging():
    respx.get("/RealTimeData").mock(
        return_value=Response(200, json=_load_json_fixture("RealTimeData_Charging"))
    )

    envoy = await _get_mock_trydan()
    data = await envoy.get_data()
    assert data is not None

    assert data.charge_state == 2
    assert data.ready_state == 0
    assert data.charge_power == 2664
    assert data.charge_energy == 0.07
    assert data.slave_error == 0
    assert data.charge_time == 105
    assert data.house_power == 0
    assert data.fv_power == 0
    assert data.paused == 0
    assert data.locked == 0
    assert data.timer == 1
    assert data.intensity == 12
    assert data.dynamic == 0
    assert data.min_intensity == 6
    assert data.max_intensity == 16
    assert data.pause_dynamic == 0
    assert data.firmware_version == "1.6.18"
    assert data.dynamic_power_mode == 2
    assert data.contracted_power == 4600



from pathlib import Path
from typing import Any

import orjson

from pytrydan import Trydan


def _fixtures_dir() -> Path:
    return Path(__file__).parent / "fixtures"


def _load_fixture(name: str) -> str:
    with open(_fixtures_dir() / name) as read_in:
        return read_in.read()


def _load_json_fixture(case_endpoint: str) -> dict[str, Any]:
    return orjson.loads(_load_fixture(case_endpoint))


async def _get_mock_trydan(update: bool = True):  # type: ignore[no-untyped-def]
    """Return a mock Trydan."""
    trydan = Trydan("127.0.0.1")
    return trydan



import pytest
import respx
from httpx import Response

from pytrydan import TrydanInvalidValue

from .conftest import _get_mock_trydan, _load_json_fixture


@pytest.mark.asyncio
@respx.mock
async def test_binary_sensors():
    respx.get("/RealTimeData").mock(
        return_value=Response(200, json=_load_json_fixture("RealTimeData"))
    )

    envoy = await _get_mock_trydan()

    envoy_data = await envoy.get_data()

    assert envoy_data is not None

    assert envoy.firmware_version == "1.6.18"

    assert envoy.connected is True
    assert envoy.charging is False
    assert envoy.ready is True

    respx.get("/write/Paused=1").mock(return_value=Response(200, text="OK"))
    assert await envoy.pause() is None

    respx.get("/write/Paused=0").mock(return_value=Response(200, text="OK"))
    assert await envoy.resume() is None

    respx.get("/write/Intensity=10").mock(return_value=Response(200, text="OK"))
    assert await envoy.intensity(10) is None

    with pytest.raises(TrydanInvalidValue) as e_info:
        await envoy.intensity(100)
        assert "Intensity must be between 6 and 32" == str(e_info.value)



KEYWORDS = {
    "ChargeState",
    "ChargePower",
    "ChargeEnergy",
    "SlaveError",
    "ChargeTime",
    "HousePower",
    "FVPower",
    "Paused",
    "Locked",
    "Timer",
    "Intensity",
    "Dynamic",
    "MinIntensity",
    "MaxIntensity",
    "PauseDynamic",
    "DynamicPowerMode",
    "ContractedPower",
}

API_TIMEOUT = 10



class TrydanError(Exception):
    """Base class for Trydan errors."""


class TrydanInvalidResponse(TrydanError):
    """Raised when Trydan returns an invalid response."""


class TrydanCommunicationError(TrydanError):
    """Raised when there is a communication error while talking to Trydan."""


class TrydanInvalidKeyword(TrydanError):
    """Raised when invalid keyword is used."""


class TrydanInvalidValue(TrydanError):
    """Raised when invalid value is used."""


class TrydanRetryLater(TrydanError):
    """Raised when value is not yet available."""



from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from typing import Any


class ChargeState(IntEnum):
    """Enum for Charge State."""

    NOT_CONNECTED = 0
    CONNECTED_NOT_CHARGING = 1
    CONNECTED_CHARGING = 2


class ReadyState(IntEnum):
    """Enum for Ready State."""

    NOT_READY = 0
    READY = 1


class SlaveCommunicationState(IntEnum):
    """Enum for Slave Communication State."""

    NO_ERROR = 0
    COMMUNICATION = 1
    READING = 2
    SLAVE = 3
    WAITING_WIFI = 4
    WAITING_COMMUNICATION = 5
    WRONG_IP = 6
    SLAVE_NOT_FOUND = 7
    WRONG_SLAVE = 8
    NO_RESPONSE = 9
    CLAMP_NOT_CONNECTED = 10
    # MODBUS_ERRORS
    ILLEGAL_FUNCTION = 21
    ILLEGAL_DATA_ADDRESS = 22
    ILLEGAL_DATA_VALUE = 23
    SERVER_DEVICE_FAILURE = 24
    ACKNOWLEDGE = 25
    SERVER_DEVICE_BUSY = 26
    NEGATIVE_ACKNOWLEDGE = 27
    MEMORY_PARITY_ERROR = 28
    GATEWAY_PATH_UNAVAILABLE = 30
    GATEWAY_TARGET_NO_RESP = 31
    SERVER_RTU_INACTIVE244_TIMEOUT = 32
    INVALID_SERVER = 245
    CRC_ERROR = 246
    FC_MISMATCH = 247
    SERVER_ID_MISMATCH = 248
    PACKET_LENGTH_ERROR = 249
    PARAMETER_COUNT_ERROR = 250
    PARAMETER_LIMIT_ERROR = 251
    REQUEST_QUEUE_FULL = 252
    ILLEGAL_IP_OR_PORT = 253
    IP_CONNECTION_FAILED = 254
    TCP_HEAD_MISMATCH = 255
    EMPTY_MESSAGE = 256
    UNDEFINED_ERROR = 257


class PauseState(IntEnum):
    """Enum for Pause State."""

    PAUSED = 1
    NOT_PAUSED = 0


class LockState(IntEnum):
    """Enum for Lock State."""

    ENABLED = 1
    DISABLED = 0


class ChargePointTimerState(IntEnum):
    """Enum for Charge Point Timer State."""

    TIMER_OFF = 0
    TIMER_ON = 1


class DynamicState(IntEnum):
    """Enum for Dynamic Intensity Modulation State."""

    DISABLED = 0
    ENABLED = 1


class PauseDynamicState(IntEnum):
    """Enum for Pause Dynamic."""

    MODULATING = 0
    NOT_MODULATING = 1


class DynamicPowerMode(IntEnum):
    """Enum for Dynamic Power Mode."""

    TIMED_POWER_ENABLED = 0
    TIMED_POWER_DISABLED = 1
    TIMED_POWER_DISABLED_AND_EXCLUSIVE_MODE_SETTED = 2
    TIMED_POWER_DISABLED_AND_MIN_POWER_MODE_SETTED = 3
    TIMED_POWER_DISABLED_AND_GRID_FV_MODE_SETTED = 4
    TIMED_POWER_DISABLED_AND_STOP_MODE_SETTED = 5


@dataclass(slots=True)
class TrydanData:
    """Model for Trydan data."""

    ID: str | None
    charge_state: int
    ready_state: int | None
    charge_power: float
    voltage_installation: int | None
    charge_energy: float
    slave_error: SlaveCommunicationState
    charge_time: int
    house_power: int
    fv_power: float
    battery_power: float | None
    paused: int
    locked: LockState
    timer: ChargePointTimerState
    intensity: int
    dynamic: DynamicState
    min_intensity: int
    max_intensity: int
    pause_dynamic: PauseDynamicState
    dynamic_power_mode: DynamicPowerMode
    contracted_power: int
    firmware_version: str | None
    SSID: str | None
    IP: str | None
    signal_status: int | None

    @classmethod
    def from_api(cls, data: dict[str, Any]) -> TrydanData:
        """Initialize from the API."""
        return cls(
            ID=data.get("ID"),
            charge_state=ChargeState(data["ChargeState"]),
            ready_state=ReadyState(data.get("ReadyState", 0)),
            charge_power=data["ChargePower"],
            voltage_installation=data.get("VoltageInstallation"),
            charge_energy=data["ChargeEnergy"],
            slave_error=SlaveCommunicationState(data["SlaveError"]),
            charge_time=data["ChargeTime"],
            house_power=data["HousePower"],
            fv_power=data["FVPower"],
            battery_power=data.get("BatteryPower"),
            paused=PauseState(data["Paused"]),
            locked=LockState(data["Locked"]),
            timer=ChargePointTimerState(data["Timer"]),
            intensity=data["Intensity"],
            dynamic=DynamicState(data["Dynamic"]),
            min_intensity=data["MinIntensity"],
            max_intensity=data["MaxIntensity"],
            pause_dynamic=PauseDynamicState(data["PauseDynamic"]),
            dynamic_power_mode=DynamicPowerMode(data["DynamicPowerMode"]),
            contracted_power=data["ContractedPower"],
            firmware_version=data.get("FirmwareVersion"),
            SSID=data.get("SSID"),
            IP=data.get("IP"),
            signal_status=data.get("SignalStatus"),
        )



from typer.testing import CliRunner

from pytrydan.cli import app

runner = CliRunner()


def test_help():
    """The help message includes the CLI name."""
    result = runner.invoke(app, ["--help"])
    assert result.exit_code == 0
    assert "Set KeyWord value in Trydan." in result.stdout
    assert "Retrieve Trydan Status." in result.stdout



__version__ = "0.0.0"

from .exceptions import (
    TrydanCommunicationError,
    TrydanInvalidKeyword,
    TrydanInvalidResponse,
    TrydanInvalidValue,
    TrydanRetryLater,
)
from .models.trydan import (
    ChargePointTimerState,
    DynamicPowerMode,
    DynamicState,
    LockState,
    PauseDynamicState,
    PauseState,
    SlaveCommunicationState,
    TrydanData,
)
from .trydan import Trydan

__all__ = [
    "Trydan",
    "TrydanData",
    "SlaveCommunicationState",
    "LockState",
    "ChargePointTimerState",
    "DynamicState",
    "PauseDynamicState",
    "PauseState",
    "DynamicPowerMode",
    "TrydanCommunicationError",
    "TrydanInvalidKeyword",
    "TrydanInvalidResponse",
    "TrydanInvalidValue",
    "TrydanRetryLater",
]
